#!/usr/bin/env python3
"""
Optimized MCP Bridge for TowardsMCP Server

Simplified bridge with minimal dependencies for connecting
MCP clients (Claude, Claude-VSCode) to TowardsMCP Server resources.
"""

import asyncio
import json
import logging
import os
import sys
from typing import Any, Dict, List, Optional

import httpx

# Set up logging to stderr (stdout is reserved for MCP communication)
logging.basicConfig(
    stream=sys.stderr,
    level=logging.WARNING,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger(__name__)

class OptimizedMCPBridge:
    """Optimized MCP-compliant bridge to TowardsMCP server"""
    
    def __init__(self, resource_name: str = None):
        # Auto-detect server URL from environment or default
        self.server_url = os.getenv('TOWARDSMCP_SERVER_URL', 'http://localhost:8081').rstrip('/')
        self.resource_name = resource_name or os.getenv('TOWARDSMCP_RESOURCE')
        self.client = None
        
        # Get API key from environment
        self.api_key = os.getenv('TOWARDSMCP_API_KEY') or os.getenv('MCP_API_KEY')
        
        if not self.api_key:
            logger.error("No API key found. Set TOWARDSMCP_API_KEY environment variable")
            sys.exit(1)
        
        if not self.resource_name:
            logger.error("Resource name is required. Set TOWARDSMCP_RESOURCE environment variable or pass as argument")
            sys.exit(1)
    
    async def initialize_client(self):
        """Initialize HTTP client for backend communication"""
        if not self.client:
            self.client = httpx.AsyncClient(timeout=30.0)
    
    async def close_client(self):
        """Close HTTP client"""
        if self.client:
            await self.client.aclose()
            self.client = None
    
    async def send_to_backend(self, method: str, params: Optional[Dict] = None) -> Dict[str, Any]:
        """Send request to TowardsMCP server backend"""
        try:
            await self.initialize_client()
            
            if method == "initialize":
                response = await self.client.get(
                    f"{self.server_url}/api/resources/{self.resource_name}/tools",
                    headers={"Authorization": f"Bearer {self.api_key}"}
                )
                
                if response.status_code == 200:
                    data = response.json()
                    return {
                        "protocolVersion": "2024-11-05",
                        "capabilities": {
                            "tools": {"listChanged": True},
                            "resources": {"listChanged": True}
                        },
                        "serverInfo": {
                            "name": f"TowardsMCP-{self.resource_name}",
                            "version": "1.0.0",
                            "description": f"TowardsMCP Bridge - {self.resource_name} Resource"
                        }
                    }
                else:
                    raise Exception(f"Failed to connect to resource: {response.text}")
            
            elif method == "notifications/initialized":
                # Handle initialization notification - no response needed for notifications
                logger.debug("Received initialization notification")
                return None  # Notifications don't require responses
            
            elif method == "prompts/list":
                # Return empty prompts list since we don't support prompts
                return {"prompts": []}
            
            elif method == "tools/list":
                response = await self.client.get(
                    f"{self.server_url}/api/resources/{self.resource_name}/tools",
                    headers={"Authorization": f"Bearer {self.api_key}"}
                )
                
                if response.status_code == 200:
                    data = response.json()
                    tools = data.get('tools', [])
                    
                    mcp_tools = []
                    for tool in tools:
                        # Ensure all required fields are present and properly typed
                        tool_name = tool.get('name')
                        tool_description = tool.get('description')
                        tool_schema = tool.get('inputSchema', {})
                        
                        # Validate required fields
                        if not tool_name or not isinstance(tool_name, str):
                            continue
                        if not tool_description or not isinstance(tool_description, str):
                            tool_description = f"Tool: {tool_name}"
                        
                        # Ensure schema is valid
                        if not isinstance(tool_schema, dict):
                            tool_schema = {"type": "object", "properties": {}}
                        
                        # Ensure schema has required type
                        if 'type' not in tool_schema:
                            tool_schema['type'] = 'object'
                        
                        # Ensure properties exists for object types
                        if tool_schema.get('type') == 'object' and 'properties' not in tool_schema:
                            tool_schema['properties'] = {}
                        
                        mcp_tools.append({
                            "name": tool_name,
                            "description": tool_description,
                            "inputSchema": tool_schema
                        })
                    
                    return {"tools": mcp_tools}
                else:
                    raise Exception(f"Failed to list tools: {response.text}")
            
            elif method == "tools/call":
                tool_name = params.get('name')
                arguments = params.get('arguments', {})
                
                if not tool_name:
                    raise Exception("Tool name is required")
                
                response = await self.client.post(
                    f"{self.server_url}/api/tools/execute",
                    headers={"Authorization": f"Bearer {self.api_key}"},
                    json={
                        "tool_name": tool_name,
                        "resource_name": self.resource_name,
                        "arguments": arguments
                    }
                )
                
                if response.status_code == 200:
                    data = response.json()
                    result_text = data.get('result', 'Tool executed successfully')
                    
                    if isinstance(result_text, (dict, list)):
                        result_text = json.dumps(result_text, indent=2)
                    elif result_text is None:
                        result_text = "Tool executed successfully (no result)"
                    else:
                        result_text = str(result_text)
                    
                    return {
                        "content": [
                            {
                                "type": "text",
                                "text": result_text
                            }
                        ]
                    }
                else:
                    error_text = response.text
                    try:
                        error_data = response.json()
                        error_text = error_data.get('detail', error_text)
                    except:
                        pass
                    raise Exception(f"Tool execution failed: {error_text}")
            
            elif method == "resources/list":
                return {
                    "resources": [
                        {
                            "uri": f"resource://{self.resource_name}",
                            "name": self.resource_name,
                            "description": f"{self.resource_name} resource",
                            "mimeType": "application/json"
                        }
                    ]
                }
            
            elif method == "resources/read":
                uri = params.get('uri', f"resource://{self.resource_name}")
                resource_name_from_uri = uri.replace("resource://", "")
                
                # Case-insensitive comparison to handle variations
                if resource_name_from_uri.lower() != self.resource_name.lower():
                    logger.error(f"Resource mismatch: requested='{resource_name_from_uri}', expected='{self.resource_name}'")
                    raise Exception(f"Resource {resource_name_from_uri} not found")
                
                # Use the tools endpoint which accepts API key auth and includes resource info
                response = await self.client.get(
                    f"{self.server_url}/api/resources/{self.resource_name}/tools",
                    headers={"Authorization": f"Bearer {self.api_key}"}
                )
                
                if response.status_code == 200:
                    data = response.json()
                    # Create a resource info object from the tools response
                    resource_info = {
                        "name": data.get("resource_name", self.resource_name),
                        "type": data.get("resource_type", "unknown"),
                        "tools": data.get("tools", []),
                        "tool_count": data.get("count", 0),
                        "description": f"{self.resource_name} resource",
                        "enabled": True  # If we can fetch tools, it's enabled
                    }
                    
                    return {
                        "contents": [
                            {
                                "uri": uri,
                                "mimeType": "application/json",
                                "text": json.dumps(resource_info, indent=2)
                            }
                        ]
                    }
                else:
                    raise Exception(f"Resource {self.resource_name} not found or not accessible")
            
            else:
                raise Exception(f"Unsupported method: {method}")
                
        except Exception as e:
            logger.error(f"Backend communication error: {e}")
            raise
    
    async def handle_request(self, request: Dict[str, Any]) -> Dict[str, Any]:
        """Handle incoming MCP request with strict protocol compliance"""
        request_id = request.get('id')
        method = request.get('method')
        
        # Check if this is a notification (no id field)
        is_notification = request_id is None
        
        # Ensure ID is properly formatted (string or number, never null for responses)
        if request_id is None and not is_notification:
            request_id = 1  # Default ID for requests without ID
        
        try:
            params = request.get('params', {})
            
            # Validate required fields
            if not method:
                if is_notification:
                    return None  # Don't respond to invalid notifications
                return {
                    "jsonrpc": "2.0",
                    "id": request_id,
                    "error": {
                        "code": -32600,
                        "message": "Invalid Request: missing method"
                    }
                }
            
            logger.debug(f"Handling method: {method} with ID: {request_id} (notification: {is_notification})")
            
            result = await self.send_to_backend(method, params)
            
            # Don't send responses for notifications
            if is_notification or result is None:
                return None
            
            return {
                "jsonrpc": "2.0",
                "id": request_id,
                "result": result
            }
            
        except Exception as e:
            logger.error(f"Error handling request: {e}")
            
            # Don't send error responses for notifications
            if is_notification:
                return None
                
            return {
                "jsonrpc": "2.0",
                "id": request_id,
                "error": {
                    "code": -32603,
                    "message": str(e)
                }
            }
    
    async def run_stdio(self):
        """Run the stdio MCP server loop"""
        logger.info(f"Starting optimized MCP bridge for resource: {self.resource_name}")
        logger.info(f"Backend server: {self.server_url}")
        
        try:
            while True:
                line = await asyncio.get_event_loop().run_in_executor(None, sys.stdin.readline)
                
                if not line:
                    break
                
                line = line.strip()
                if not line:
                    continue
                
                try:
                    request = json.loads(line)
                    logger.debug(f"Received request: {request}")
                    
                    response = await self.handle_request(request)
                    
                    # Only send response if it's not None (notifications don't get responses)
                    if response is not None:
                        logger.debug(f"Sending response: {response}")
                        print(json.dumps(response), flush=True)
                    else:
                        logger.debug("No response needed (notification)")
                    
                except json.JSONDecodeError as e:
                    logger.error(f"Invalid JSON received: {e}")
                    error_response = {
                        "jsonrpc": "2.0",
                        "id": 1,  # Use default ID instead of null
                        "error": {
                            "code": -32700,
                            "message": "Parse error"
                        }
                    }
                    print(json.dumps(error_response), flush=True)
                
                except Exception as e:
                    logger.error(f"Error processing request: {e}")
                    # Try to extract ID from the request if possible
                    request_id = 1
                    try:
                        if 'request' in locals() and isinstance(request, dict):
                            request_id = request.get('id', 1)
                            if request_id is None:
                                request_id = 1
                    except:
                        pass
                    
                    error_response = {
                        "jsonrpc": "2.0",
                        "id": request_id,
                        "error": {
                            "code": -32603,
                            "message": f"Internal error: {str(e)}"
                        }
                    }
                    print(json.dumps(error_response), flush=True)
        
        except KeyboardInterrupt:
            logger.info("Shutting down MCP bridge")
        
        finally:
            await self.close_client()

def main():
    """Main entry point with simplified arguments"""
    import argparse
    
    parser = argparse.ArgumentParser(description="Optimized MCP Bridge for TowardsMCP Server")
    parser.add_argument('resource_name', nargs='?', help='Resource name to connect to')
    parser.add_argument('--debug', action='store_true', help='Enable debug logging')
    
    args = parser.parse_args()
    
    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)
    
    # Create and run bridge
    bridge = OptimizedMCPBridge(resource_name=args.resource_name)
    
    # Run the stdio loop
    asyncio.run(bridge.run_stdio())

if __name__ == "__main__":
    main() 